module Ball(
  Ball,
  initBall,
  lookAtBall,
  increaseSpeed,
  moveBall,
  checkBrickCollisions,
  checkPaddleCollision
) where

import Brick
import GameConstants
import Paddle
import Util

import qualified Graphics.UI.GLUT as GL
import Data.Maybe

data Ball = Ball {
  position :: Position,
  travelling :: Vector,
  speed :: Speed,
  radius :: Size
}

initBall :: Ball
initBall = Ball {
  position   = Position (gameWidth / 2) (gameHeight / 2),
  travelling = normalize $ Vector (0.2) (-0.5),
  speed      = 1.5,
  radius     = 2.5
}

speedIncreaseFactor :: Double
speedIncreaseFactor = 1.001

instance Renderable Ball where
  render b = GL.preservingMatrix $ do
    translateTo . position $ b
    GL.renderObject GL.Wireframe (GL.Cube (radius b))

lookAtBall :: Ball -> IO ()
lookAtBall b = GL.lookAt cameraPosition cameraLookingAt cameraUp
  where
    (Position ballX ballY)  = position b
    cameraPosition  = GL.Vertex3 (gameWidth / 2) (gameHeight / 2) (cameraDepthOffset)
    cameraLookingAt = GL.Vertex3 ballX ballY 0
    cameraUp        = GL.Vector3 0 1 0

increaseSpeed :: Ball -> Ball
increaseSpeed ball = ball {
    speed = (speed ball) * speedIncreaseFactor
}

moveBall :: Ball -> BallMoveResult
moveBall b = checkWallCollision b {
    position = (position b) .+. ((speed b) .*. (travelling b))
}
  where
    checkWallCollision :: Ball -> BallMoveResult
    checkWallCollision = checkBottom . checkLeft . checkRight . checkTop
      where
        checkLeft   = checkX (>) (+) 0 id id
        checkRight  = checkX (<) (-) gameWidth id id
        checkTop    = checkY (<) (-) gameHeight id id
        checkBottom = checkY (>) (+) 0 BallOk BallHitBottom
        checkX cmp fn base = check cmp base x (\n p -> p { x = n }) i (\n v -> v { i = n }) fn
        checkY cmp fn base = check cmp base y (\n p -> p { y = n }) j (\n v -> v { j = n }) fn

        check comparision base getValue setValue getComponent setComponent changeFn defaultFn changedFn ball 
          | (base `changeFn` (radius ball)) `comparision` (getValue . position $ ball) = changedFn res
          | otherwise = defaultFn ball
          where
            res = ball {
              position = updatedPosition,
              travelling = updatedTravelling
            }
            updatedPosition = setValue (base `changeFn` (radius ball)) (position ball)
            updatedTravelling = setComponent (negate . getComponent . travelling $ ball) (travelling ball)

checkBrickCollisions :: Ball -> [Brick] -> BallBrickCollisionResults
checkBrickCollisions ball = foldr checkBrickCollision initBBCR
  where
    initBBCR = BBCR { newBall = ball, remainingBricks = [], killedBricks = [] }

    checkBrickCollision :: Brick -> BallBrickCollisionResults -> BallBrickCollisionResults
    checkBrickCollision brick bbcr
      | isPositionInsideBrick top     brick || 
        isPositionInsideBrick bottom  brick = bbcr { 
            newBall = ball { travelling = flippedYTravelling },
            killedBricks = brick:(killedBricks bbcr)
          }
      | isPositionInsideBrick left    brick || 
        isPositionInsideBrick right   brick = bbcr {
            newBall = ball { travelling = flippedXTravelling },
            killedBricks = brick:(killedBricks bbcr)
          }
      | otherwise                           = bbcr {
            remainingBricks = brick:(remainingBricks bbcr)
          }
      where
        ball = newBall bbcr
        pos@(Position xp yp) = position ball
        top     = pos { y = yp + (radius ball) }
        bottom  = pos { y = yp - (radius ball) }
        left    = pos { x = xp - (radius ball) }
        right   = pos { x = xp + (radius ball) }
        flippedXTravelling = (travelling ball) { i = negate . i . travelling $ ball }
        flippedYTravelling = (travelling ball) { j = negate . j . travelling $ ball }

checkPaddleCollision :: Paddle -> Ball -> Ball
checkPaddleCollision paddle ball = fromMaybe ball $ do
  vector <- calculateReturnVector paddle bottom
  return $ ball { travelling = vector }
  where
    pos@(Position _ yp) = position ball
    bottom  = pos { y = yp - (radius ball) }
